"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.reflect = void 0;
const decorators_1 = require("./decorators");
const helpers_1 = require("./helpers");
const parser_1 = require("./parser");
const walker_1 = require("./walker");
function reflectClass(target) {
    const visitorOrder = [
        walker_1.visitors.addsDesignTypes,
        walker_1.visitors.addsDecorators,
        walker_1.visitors.addsTypeOverridden,
        walker_1.visitors.addsParameterProperties,
        walker_1.visitors.addsGenericOverridden,
        walker_1.visitors.addsTypeClassification,
        walker_1.visitors.removeIgnored,
    ];
    const visitor = (value, ctx) => visitorOrder.reduce((a, b) => !!a ? b(a, ctx) : a, value);
    return walker_1.walkClass(target, { visitor, classPath: [] });
}
function traverseObject(fn, name, ctx) {
    if (fn === undefined || fn === null)
        return;
    if (Array.isArray(fn))
        return;
    if (typeof fn === "object") {
        // CIRCULAR: return immediately
        if (ctx.path.some(x => x === fn))
            return;
        return reflectObject(fn, name, { path: ctx.path.concat(fn) });
    }
    if (typeof fn === "function" && helpers_1.metadata.isConstructor(fn))
        return reflectClass(fn);
    if (typeof fn === "function")
        return parser_1.parseFunction(fn);
}
function reflectObject(object, name, ctx) {
    return {
        kind: "Object", name,
        members: Object.keys(object).map(x => traverseObject(object[x], x, ctx)).filter((x) => !!x)
    };
}
function reflectModuleOrClass(opt) {
    if (typeof opt === "string") {
        return reflectObject(require(opt), "module", { path: [] });
    }
    else {
        return reflectClass(opt);
    }
}
const cacheStore = new Map();
const reflectCached = helpers_1.useCache(cacheStore, reflectModuleOrClass, x => x);
function reflect(pathOrClass, opt) {
    if (opt === null || opt === void 0 ? void 0 : opt.flushCache)
        cacheStore.delete(pathOrClass);
    return reflectCached(pathOrClass);
}
exports.reflect = reflect;
// --------------------------------------------------------------------- //
// ----------------------------- DECORATORS ---------------------------- //
// --------------------------------------------------------------------- //
/**
 * Decorator that do nothing, intended to be able to inspect data type
 */
reflect.noop = decorators_1.noop;
/**
 * Ignore member from metadata generated
 */
reflect.ignore = decorators_1.ignore;
/**
 * Override type definition information. Useful to add type definition for some data type that is erased
 * after transfile such as Partial<Type> or ReadOnly<Type>
 *
 * If applied to parameter it will override the parameter type
 *
 * If applied to property it will override the property type
 *
 * if applied to method it will overrid the method return value
 * @param type The type overridden
 * @param info Additional information about type (readonly, partial etc)
 */
reflect.type = decorators_1.type;
/**
 * Create class dynamically
 */
reflect.create = helpers_1.createClass;
/**
 * Mark all constructor parameters as properties
 */
reflect.parameterProperties = decorators_1.parameterProperties;
